#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <math.h>

#define LINE_LEN 32768
// #define MAX_VALS 131072
#define MAX_VALS 1600
#define MAX_COLS 200
int *hist[MAX_COLS];
int threshold[MAX_COLS];

int main(int argc, char **argv)
{
    FILE *ifp, *ofp;
    char *spc;
    int i, col, val, outval, intervalWidth;
    int sum, nvals;
    char lineStr[LINE_LEN];

    if (argc < 3) 
    {
        printf("Usage: %s <input file> <output file> <interval width>\n", argv[0]);
        exit(0);
    }

    if ((ifp = fopen(argv[1], "r")) == NULL) 
    {
        printf("%s doesn't exist\n", argv[1]);
        exit(0);
    }

    if ((ofp = fopen(argv[2], "w")) == NULL) 
    {
      	printf("Unable to open %s for writing\n", argv[2]);
        exit(0);
    }

    intervalWidth = atoi(argv[3]);

    for (i = 0; i < MAX_COLS; i++) 
    {
      hist[i] = calloc(MAX_VALS, sizeof(int));
    }

    printf("Done with allocating memory...\n");

    // compute the histogram 
    while (!feof(ifp)) 
    {
        if (fgets(lineStr, LINE_LEN, ifp) == NULL)
            break;
      	col = 0;
	// printf("Read: %s\n", lineStr);
	spc = lineStr;
	while (1)
	{
	  while (isspace(*spc))
	  {
	    ++spc;
	  }
	  if ((*spc == '\0') || (*spc == '\n'))
	    break;
	  val = atoi(spc);
	  // group them in intervals of width 100
	  val /= intervalWidth;
	  // printf("Looking at %d\n", val);
	  hist[col][val]++;
	  // skip over the # we just read
	  while (isdigit(*spc))
         {
	    ++spc;
	  }
	  ++col;
	  if (col >= MAX_COLS)
	    break;
	}
	fflush(stdout);
    }

    // for each column, compute the average # of entries in each
    // bucket of the histogram.  then, to spread the range of values,
    // if the # of entries in a bucket is below the average, output
    // the value as is; otherwise, output log2() of the histogram.
    for (i = 0; i < col; ++i)
    {
      sum = nvals = 0;
      for (val = 0; val < MAX_VALS; ++val)
      {
	if (hist[i][val] == 0)
	  continue;
	sum += hist[i][val];
	nvals++;
      }
      if (nvals > 0)
	threshold[i] = sum / nvals;
      else
	threshold[i] = 0;
      printf("Threshold for col=%d is %d\n", i, threshold[i]);
    }

    // output the binned values
    fseek(ifp, 0, SEEK_SET);
    while (!feof(ifp))
    {
	int histval;

      	col = 0;
        if (fgets(lineStr, LINE_LEN, ifp) == NULL)
            break;
	spc = lineStr;
	while (1)
	{
	  while (isspace(*spc))
	 {
	    ++spc;
	  }
	  if ((*spc == '\0') || (*spc == '\n'))
	    break;
	  val = atoi(spc);
	  histval = hist[col][val / intervalWidth];
	  if (histval > threshold[col])
	  	outval = (int) round(log2((double) histval));
	  else
	  	outval = val;
	  
	  fprintf(ofp, "%d ", outval);
	  // skip over the # we just read
	  while (isdigit(*spc))
	  {
	    ++spc;
	  }
	  ++col;
	  if (col >= MAX_COLS)
	    break;
	}
	fflush(stdout);
	fprintf(ofp, "\n");
	fflush(ofp);
    }
    fclose(ofp);
    fclose(ifp);
}
