
#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

float ran1(long *idum)
/* generate a unifrom random deviate between 0.0 and 1.0 */
{
    int j;
    long k;
    static long iy = 0;
    static long iv[NTAB];
    float temp;

    if ((*idum <= 0) || !iy) {
        if (-(*idum) < 1) *idum = 1;
        else
            *idum = -(*idum);
        for (j = NTAB + 7; j>=0; j--) {
            k = (*idum)/IQ;
            *idum = IA * (*idum -k * IQ) - IR * k;
            if (*idum < 0)
                *idum += IM;
            if (j < NTAB) iv[j] = *idum;
        }
        iy = iv[0];
    }
    k = (*idum)/IQ;
    *idum = IA*(*idum-k*IQ)-IR*k;
    if (*idum < 0) *idum += IM;
    j = iy/NDIV;
    iy = iv[j];
    iv[j] = *idum;
    if ((temp=AM*iy) > RNMX) return RNMX;
    return temp;
}

float expdev(long *idum)
/* return an exponentially distributed positive random deviate of unit mean */
{
    float dum;
    
    do
        dum = ran1(idum);
    while (dum == 0.0);
    return -log(dum);
}

int main(int argc, char **argv)
{
    long seed = -1;
    long num;
    float rnum;
    int scale = 5000, i, j, numLines, numValsPerLine;

    if (argc < 4) {
        printf("Usage: %s <scale> <#of lines> <# of vals per line>\n", argv[0]);
        exit(0);
    }
    scale = atoi(argv[1]);
    numLines = atoi(argv[2]);
    numValsPerLine = atoi(argv[3]);
    for (i = 0; i < numLines; i++) {
        for (j = 0; j < numValsPerLine;j++) {
            rnum = expdev(&seed);
            rnum *= scale;
            num = lroundf(rnum);
            printf(" %u ", num);
        }
        printf("\n");
    }
}
