
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#define MAX_VALUES 1000
long zipf_freq_dist[MAX_VALUES];

void generate_zipf_freq(float alpha, float beta);
void output_zipf_dist(int num_vals_per_line);
int vals_exist_for_output(void);

int main(int argc, char **argv)
{
	float alpha, beta;;
	int i, j, num_vals_per_line;
	
	if (argc < 4) 
	{
		printf("Usage: %s <alpha> <beta> <num_vals_per_line of output>\n", argv[0]);
		exit(0);
	}

	alpha = atof(argv[1]);
	beta = atof(argv[2]);
	num_vals_per_line = atoi(argv[3]);
	
	generate_zipf_freq(alpha, beta);
	output_zipf_dist(num_vals_per_line);
}

/*
 * Generate zipf freq.
*/
void generate_zipf_freq(float alpha, float beta)
{
	int i;

	for (i = 1; i < MAX_VALUES; i++) 
        {
		zipf_freq_dist[i] = lroundf(beta / pow(i, alpha));
printf("%d %ld\n", i, zipf_freq_dist[i]);
	}
exit(0);
}

void output_zipf_dist(int num_vals_per_line)
{
  int i, j;

	while (1) {
	  i = 0;
	  while (i < num_vals_per_line) {
	    if (!vals_exist_for_output()) {
	      printf("\n");
	      return;
	    }
	    j = (lrand48()) % MAX_VALUES;
	    if (zipf_freq_dist[j] > 0) {
	      printf("%d ", j);
	      i++;
	      zipf_freq_dist[j]--;
	    }
	  }
	  printf("\n");;
	}
}

int vals_exist_for_output()
{
  int k, vals_exist = 0;

  for (k = 0; k < MAX_VALUES; k++)
  {
    if (zipf_freq_dist[k] > 0)
    {
      vals_exist = 1;
      break;
    }
  }
  return vals_exist;
}
