/**

PROJECT:		CS 297 Deliverable 2

FILE:			AttrValLattice.java

PURPOSE:		Constructs bitmaps and then constructs the Attribut Value Lattice


@author  Aarathi Raghu

@version	12/05/2005

@since	Java 1.4.2

		
*/

import java.util.*;
import java.lang.*;

public class AttrValLattice 
{
	int lineLength=0;
	ByteClass latticeBitmap[];
	TreeSet sortedNodeSet;
	HashSet closedFrequentItemSet;
	int numCFsets = 0;
    	final static int latticeBitmapSize = 70000;
	/**
	constructor
	declare a lattice bitMap and a sorted Node Set	
	@param numCols: The # of columns in the input data that corresponds to items in a transaction
	*/
	public AttrValLattice(int numCols) 
	{
            latticeBitmap = new ByteClass[latticeBitmapSize];
		sortedNodeSet = new TreeSet( new NodeInfo() );
		lineLength = numCols;
		closedFrequentItemSet = null;
		numCFsets = 0;
	}
	
	/**
	   Construct the bitmap for each column of the input (i.e., construct the granules)
		@param inputFile : This file reads in an inputFile and constructs a bitmap for each value
		@param numLines : The # of lines in the data (corresponds to the # of transactions)
	*/
	public void MakeBitmap(String inputFile,int numLines) throws Exception
	{
		DiskReader dr = new DiskReader(inputFile);
		int values [] = new int [lineLength];
		int numValsInLine, lineNum = 0;
		Object mapValue;
		ByteClass bc;

		while (true) 
		{
			if ((numValsInLine = dr.ReadLine(values)) < 0)
				break;

			lineNum++;

			for (int i = 0; i < numValsInLine; i++) 
			{
				// if there is a bitmap already for values[i], update it to reflect that
				// "intVal" occurs in the current line.

                            	if (latticeBitmap[values[i]] == null)
				{
				    // 1 byte can represent 8 lines of input.  So, size accordingly
                                    latticeBitmap[values[i]]  = new ByteClass(numLines);
				}
                                latticeBitmap[values[i]].set(lineNum);

			}
		}

                System.out.println("Read of: " + inputFile + " is complete ");
	}
	/**
	 @param minSup: This is the minimum Support we want the algorithm to use
	 Go thru the keys of the latticeBitmap and construct nodes for which the cardinality
	 of the bitmap is above the minSup.
	*/
	public void MakeNodes(int minSup)
	{
            Integer key = new Integer(0);
            NodeInfo n;
            HashSet keySet;
            int count = 0;

            for (int i = 0; i < latticeBitmapSize; i++)
	    {
                if ((latticeBitmap[i] == null) ||
                    (latticeBitmap[i].cardinality() <= minSup))
                    continue;
                key = i;
			
                keySet = new HashSet();
                keySet.add(key);
                // put this in a list
                n = new NodeInfo(keySet, latticeBitmap[i]);
                sortedNodeSet.add(n);
                count++;
                latticeBitmap[i] = null;
            }
            // don't need it anymore.  can release this to free memory
            latticeBitmap = null;
	}
	
	/**
	@param minSup : the minimum support to use in the algorithm
	Based on phase 1 of th ealgorithm prescribed in Lin, Hu, and Louie's paper
	Implement algorithm #1 to construct Attribute-value lattice
	*/

	public void CombineNodes(int minSup)
	{
		Iterator sortedNodeIterator = sortedNodeSet.iterator(), siblingSetIterator;
		TreeSet siblingSet;
		NodeInfo parentNode, siblingNode;
		int combinedCount;
		NodeInfo combinedNode;

		while (true) {
			boolean reset = false;
		
			if (!sortedNodeIterator.hasNext())
				break;
				
			parentNode = (NodeInfo) sortedNodeIterator.next();
			siblingSet = (TreeSet) sortedNodeSet.tailSet(parentNode);
			siblingSetIterator = siblingSet.iterator();
			while (siblingSetIterator.hasNext()) 
			{
					siblingNode = (NodeInfo) siblingSetIterator.next();
					HashSet keysUnion = UnionKeySets(parentNode.keySet, siblingNode.keySet);
					ByteClass bc = (ByteClass) parentNode.bitmap.clone();

					if (siblingNode.isVisited == true)
						continue;
					if (parentNode.equals(siblingNode))
						continue;
					if (parentNode.isDeleted == true)
						continue;
					if (siblingNode.isDeleted == true)
						continue;
						
					bc.and(siblingNode.bitmap);
					if (bc.cardinality() <= minSup)
						continue;
					if (parentNode.bitmap.equal(siblingNode.bitmap)) 
					{
						// Case I: B(I_i) == B(I_j)
						combinedNode = new NodeInfo(keysUnion, parentNode.bitmap);
						combinedNode.level = parentNode.level;
						
						parentNode.isDeleted = true;
						siblingNode.isDeleted = true;
						
						sortedNodeSet.add(combinedNode);
						
						reset = true;
						break;
					} 
					else if (parentNode.bitmap.subset(siblingNode.bitmap)) 
					{
						// Case II: B(I_i) \subset B(I_j)
						siblingNode.isVisited = true;
						if (!parentNode.edgeExists(siblingNode)) 
						{
							parentNode.addEdge(siblingNode);
							siblingNode.setMaxLevel(parentNode.level + 1);
						}
					} 
					else if (siblingNode.bitmap.subset(parentNode.bitmap)) 
					{
						// Case III: B(I_j) \subset B(I_i)
						siblingNode.isVisited = true;
						if (!siblingNode.edgeExists(parentNode))
						 {
							siblingNode.addEdge(parentNode);
							parentNode.setMaxLevel(siblingNode.level + 1);
						}
					}
			}
			if (reset == true) 
			{
				sortedNodeIterator = sortedNodeSet.iterator();
			}
			// Clear the isVisited flag for the next iteration.
			siblingSetIterator = siblingSet.iterator();
			while (siblingSetIterator.hasNext()) 
			{
				siblingNode = (NodeInfo) siblingSetIterator.next();
				if (siblingNode.isDeleted == true)
					continue;
				siblingNode.isVisited = false;
			}
			
		}
		
	}
	
	/**
	This is a simple helper function to print out the nodes of a given node set 
	@param nodeSet: nodeset neede to be printed
	*/
	public void PrintNodes(AbstractSet nodeSet)
	{
		Iterator sortedNodeIterator = nodeSet.iterator();
		NodeInfo currentNode;
		
		while (sortedNodeIterator.hasNext()) 
		{
			currentNode = (NodeInfo) sortedNodeIterator.next();
			if (currentNode.isDeleted == true)
				continue;
			currentNode.print();
		}
	}
	
	/**
	Implement algorithm #2 to find closed frequent itemsets from the attribute-value lattice.
	@param minSup: The minimum support used for finding the closed frequent itemsets
	*/
	public void FindClosedFrequentItemsets(int minSup)
	{
		Iterator sortedNodeIterator = sortedNodeSet.iterator();
		TreeSet nodeSet;
		NodeInfo currentNode;
		
		nodeSet = new TreeSet( new NodeInfo() );
		/* build the greatest lower bound items: i.e., nodes at level=1 */
		// System.out.println("Level-1 nodes are: " + sortedNodeSet.size());
		// Code for line 1 of the algorithm
		while (sortedNodeIterator.hasNext()) 
		{
			currentNode = (NodeInfo) sortedNodeIterator.next();
			if (currentNode.level != 1)
				continue;
			if (currentNode.isDeleted == true)
				continue;
			nodeSet.add(currentNode);
			currentNode.isInNewNodeSet = true;
		}
		System.out.println("level-1 nodes:" + nodeSet.size());		
		System.out.println("Starting expansion..");
		closedFrequentItemSet = new HashSet();
		// line 2: compute the frequent itemsets from the lattice
		ExpandItemSets(nodeSet, minSup);

		// line 9: from the frequent itemsets, identify the closed frequent itemsets
		GetClosedFrequentItemSets();

	}


	/**
	Expand the lattice to find frequent itemsets.
	@param nodeSet:  The candidate frequent itemsets to be considered for expansion.
	@param minSup: The minimum support used for finding the closed frequent itemsets
	*/	
	private void ExpandItemSets(TreeSet nodeSet, int minSup)
	{
		Iterator sortedNodeIterator = sortedNodeSet.iterator(), siblingSetIterator;
		TreeSet newNodeSet, siblingSet;
		NodeInfo currentNode, siblingNode, combinedNode;

		// Start with the nodes at level-1 and expand.
		
		while (sortedNodeIterator.hasNext())
		{
		    currentNode = (NodeInfo) sortedNodeIterator.next();
		    currentNode.isInNewNodeSet = false;
		}

		sortedNodeIterator = sortedNodeSet.iterator();
		/* identify the closed frequent item sets */
		sortedNodeIterator = nodeSet.iterator();
		// loop corresponding to line 2 of the algorithm
		while (sortedNodeIterator.hasNext())
		{
			newNodeSet = null;
			currentNode = (NodeInfo) sortedNodeIterator.next();
			currentNode.isInNewNodeSet = false;
			if (currentNode.level < 1)
				continue;
			
			siblingSet = (TreeSet) nodeSet.tailSet(currentNode);
			siblingSetIterator = siblingSet.iterator();
			
			// Loop corresponding to line 3 of the algorithm
			while (siblingSetIterator.hasNext()) 
			{
				siblingNode = (NodeInfo) siblingSetIterator.next();
				siblingNode.isInNewNodeSet = false;
				
				if (siblingNode.level < 1)
					continue;
				if (currentNode.equals(siblingNode))
					continue;
				
				HashSet keysUnion = UnionKeySets(currentNode.keySet, siblingNode.keySet);
				ByteClass bc = (ByteClass) currentNode.bitmap.clone();
				
				if (newNodeSet == null) {
					newNodeSet = new TreeSet( new NodeInfo() );
				}
				bc.and(siblingNode.bitmap);
				/* line 5 of the algorithm */
				if (bc.cardinality() > minSup)
				{
					combinedNode = new NodeInfo(keysUnion, bc);
					combinedNode.isInNewNodeSet = true;
					// Our addition: need to define the parents for the combined node.
					UnionEdges(combinedNode, currentNode, siblingNode);
					newNodeSet.add(combinedNode);					
				} 
				else 
				{
					// Add currentNode's parents to the NewNode set
					Iterator parentsIter = currentNode.edges.iterator();
					NodeInfo parentNode;
					while (parentsIter.hasNext()) 
					{
						parentNode = (NodeInfo) (parentsIter.next());
						if (parentNode.isDeleted == true)
							continue;
						if (!parentNode.isInNewNodeSet) {
							parentNode.isInNewNodeSet = true;
							newNodeSet.add(parentNode);
						}
					}
					
				}
				bc = null; // help the gc
			}
			if ((newNodeSet != null) && (newNodeSet.size() > 0))
			{
				/* line 6 of the algorithm */
				ExpandItemSets(newNodeSet, minSup);
			}
			// line 7/8 of the algorithm: filter the output 
			if (!isCFItemSet(currentNode))
			{
				closedFrequentItemSet.add(currentNode);
			}
		}
	}

	/**
	From the set of frequent itemsets, construct the set of closed frequent itemsets.
	*/
    	private void GetClosedFrequentItemSet()
    	{
	    	HashSet frequentItemSet = closedFrequentItemSet;
		Iterator freqSetIter = frequentItemSet.iterator();
		NodeInfo currentNode;
		HashSet compactedSet;

		closedFrequentItemSet = new HashSet();
		
		while (freqSetIter.hasNext()) 
		{
		    currentNode = (NodeInfo) freqSetIter.next();
		    FindAncestorSet(currentNode);
		}
	}

	/**
	Given a node in a lattice, recursively walk the node's parents and collect
	the set of nodes that are the parents of the node.
	@param node:   The node whose ancestors we are trying to determine
	*/		    
    	private void FindAncestorSet(NodeInfo node)
	{
	    	Iterator parentsIter = node.edges.iterator();
		NodeInfo combinedNode, parentNode;

		if (!parentsIter.hasNext()) {
		    if (!isCFItemSet(node)) {
			removeSubsumedCFItemSets(node);
			closedFrequentItemSet.add(node);
		    }
		    return;
		}

		while (parentsIter.hasNext()) 
		    {
			parentNode = (NodeInfo) (parentsIter.next());
			HashSet keysUnion = UnionKeySets(node.keySet, parentNode.keySet);
			ByteClass bc = (ByteClass) node.bitmap.clone();
			bc.and(parentNode.bitmap);
			
			combinedNode = new NodeInfo(keysUnion, bc);
			combinedNode.edges = parentNode.edges;
			FindAncestorSet(combinedNode);
		    }

	}


	/**
	This function walks the cf-itemset and returns true if node is subsumed; otherwise returns false. 
	@param node: node is already in the closed frequent itemset if it is subsumed by 
		one of the nodes in the closed frequent itemset.  Subsumed means that the
		node's bitmap is a subset of a node that is a closed frequent itemset.
	 */
	private boolean isCFItemSet(NodeInfo node)
	{
		Iterator cfSetIter = closedFrequentItemSet.iterator();
		NodeInfo currentNode;
		
		while (cfSetIter.hasNext())
		{
			currentNode = (NodeInfo) cfSetIter.next();
			if (node.equals(currentNode))
			    return true;
			if (currentNode.subsumes(node))
			{
				return true;
			}
		}
		return false;
	}
	
	/**
	  @param parentNode:				
	  We will be adding parentNode to the closed frequent item set.  As a result of the addtion,
	  remove any child nodes from the closed frequent itemset that are subsumed by parent.
	*/
	private void removeSubsumedCFItemSets(NodeInfo parentNode)
	{
		Iterator cfSetIter = closedFrequentItemSet.iterator();
		NodeInfo currentNode;
		
		while (cfSetIter.hasNext())
		{
			currentNode = (NodeInfo) cfSetIter.next();
			if (parentNode.equals(currentNode))
				continue;
			if (parentNode.subsumes(currentNode))
			{
			
				closedFrequentItemSet.remove(currentNode);
				/* reset the iterator as the list changed */
				cfSetIter = closedFrequentItemSet.iterator();
			}
		}
		
	}
	
	/**
	Set the parents of a node to be the union of two nodes.
	@param a, b: The nodes whose set of edges we are trying to union
	@param result:  The parents of result are the union of the parents of a,b
	*/
    	private void UnionEdges(NodeInfo result, NodeInfo a, NodeInfo b)
    	{
		Iterator edgesIter = a.edges.iterator();
		NodeInfo n;
	
		while (edgesIter.hasNext())
		{
	    		n = (NodeInfo) (edgesIter.next());
	    		result.addEdge(n);
		}
		edgesIter = b.edges.iterator();
		while (edgesIter.hasNext())
		{
	    		n = (NodeInfo) (edgesIter.next());
	    		if (!result.edgeExists(n))
			result.addEdge(n);
		}
	
    	}

	/**
	combine currentNode and siblingNodes's ancestors and union them to put them in the
	closed frequent itemsets
	@param currentNode : This is the current Node (belongs at level 1)
	@param siblingNode: again belongs at level 1
	@param cfSet:  The set of closed frequent sets we have constructed so far.
	
	*/
	public void AddAncestorSets(NodeInfo currentNode, NodeInfo siblingNode, HashSet cfSet)
	{
		NodeInfo parentNode, n, temp;
		HashSet keysUnion;
		ByteClass bc;
		Iterator cfIter = cfSet.iterator();
		boolean found = false;

		keysUnion = UnionKeySets(currentNode.keySet, siblingNode.keySet);
		bc = (ByteClass) currentNode.bitmap.clone();
		bc.and(siblingNode.bitmap);
		n = new NodeInfo(keysUnion, bc);
		while (cfIter.hasNext())
		{
			temp = (NodeInfo) cfIter.next();
			if (temp.equals(n)) 
			{
				found = true;
				break;
			}
		}
		if (!found)
		{
			System.out.println("Adding ancestor: ");
			n.print();
			cfSet.add(n);
		}

		if (currentNode.edges == null)
			return;
			
		// Add currentNode's parents to the NewNode set
		Iterator parentsIter = currentNode.edges.iterator();
				
		while (parentsIter.hasNext())
		{
			parentNode = (NodeInfo) (parentsIter.next());
			
			AddAncestorSets(parentNode, siblingNode, cfSet);
		}

	}
	
	/**
	Union two keysets and return the union as result.
	@param key1:one of pair of keysets to perform union operation 
	@param key2:one of pair of keysets to perform union operation
	*/
	public HashSet UnionKeySets(HashSet key1, HashSet key2)
	{
		HashSet result;
		Iterator iter = key1.iterator();
		Integer val;
		
		result = new HashSet();

		while (iter.hasNext()) 
		{
			val = (Integer) iter.next();
			result.add(new Integer(val.intValue()));
		}
		
		iter = key2.iterator();
		
		while (iter.hasNext()) 
		{
			val = (Integer) iter.next();
			result.add(new Integer(val.intValue()));
		}
				
		return result;
	}

}
