/**

PROJECT:		CS 297 Deliverable 1

FILE:			ByteClass.java

PURPOSE:		Implementation of the bitInterface.  Provides an implementation
			for fast bit manipulations.  

@author  Aarathi Raghu

@version        6/05/2005

@since  Java 1.4.2


*/



public class ByteClass implements BitInterface 
{
	short bs[];
	int bsSize;
	int nBitsSet;
	final static int MAX_POSITIONS_TRACKED = 16;
	int positionsSet[];
	int lastPosition;
	
	/**
	lookup table used for cardinality
	source : http://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetTable
	*/

	int BitsSetTable256[] = 
		{
			0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4, 
			1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
			1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
			1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
			3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
			4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
		};
		
	/**
	Constructor
	@param max:  The maximum # of bits in the bitset.
	*/	
	public ByteClass(int max) 
	{

		bsSize = max;
		positionsSet = null;
		lastPosition = 0;
		
		bs = new short[bsSize];
		for(int i = 0 ; i < bsSize ; i++)
		{
			bs[i] = 0;
		}
		// bs = null;
		nBitsSet = 0;
	}
	/**
	Compute the cardinality of the bitset (i.e., # of bits that are set to 1)
	@retval The # of bits set to 1.
	*/	
	public int cardinality() 
	{
	    return nBitsSet;
	}
	
	/**
	Set a specific bit.
	@param num  The position of the bit that should be set.
	*/	
	public void set(int num) 
	{
		if ((bs == null) && (lastPosition < MAX_POSITIONS_TRACKED)) 
		{
			if (get(num) != 0)
				return;
				
			positionsSet[lastPosition] = num;
			lastPosition++;
			nBitsSet++;
			return;
		}
		
		int loc = num >> 3; // divide by 8
		
		int off = num & 0x7; // num mod 8---get the last 3 bits
		
		// is it already set
		if (get(num) != 0)
			return;
		if (bs == null) 
		{
			convertRep();
		}
			
		bs[loc] |= ((short) (1 << off));
		nBitsSet++;
		
	}
	
	/**
	Convert to a packed representation from a sparse representation.
	*/	
	public void convertRep()
	{
		bs = new short[bsSize];
		for(int i = 0 ; i < bsSize ; i++)
		{
			bs[i] = 0;
		}
		nBitsSet = 0;
		for (int i = 0; i < lastPosition; i++) 
		{
			set(positionsSet[i]);
		}
		positionsSet = null; // help GC
		lastPosition = 0;
	}
	
	/**
	Return the value of the bit at a particular position.	
	*/		
	public int get(int num)
	{
		if (positionsSet != null) {
			for (int i = 0; i < lastPosition; i++) 
			{
				if (positionsSet[i] == num)
					return 1;
			}
			return 0;
		}
		
		int loc = num >> 3; // divide by 8
		int off = num & 0x7; // num mod 8---get the last 3 bits
		// return 1 if the bit is set; 0 otherwise
		return ((bs[loc] & ((short) (1 << off))) != 0) ? 1 : 0;
	}
	
	/**
	Perform the "and" operation on 2 bitsets.
	@param b:  The bitset with which "this" bitset should be and'ed with.
	*/		
	public void and(BitInterface b) 
	{
		ByteClass other = (ByteClass) b;
		
		if (positionsSet != null)
			convertRep();
		if (other.positionsSet != null)
			other.convertRep();
			
		nBitsSet = 0;
		for (int i = 0; i < bsSize; i++) 
		{
			if (bs[i] != 0) {
				bs[i] = (short) (bs[i] & other.bs[i]);
				nBitsSet += BitsSetTable256[bs[i]];
			}
		}
	}
	
	/**
	Perform the "equal" operation on 2 bitsets.
	@param b:  The bitset with which "this" bitset should be compared with for equality.
	@retval true if the bitsets are equal; false otherwise.
	*/		
	public boolean equal(BitInterface b)
	{
		ByteClass other = (ByteClass) b;
				
		if (positionsSet != null)
			convertRep();
		if (other.positionsSet != null)
			other.convertRep();

		for (int i = 0; i < bsSize; i++) 
		{
			if (bs[i] != other.bs[i])
				return false;
		}
		return true;
	}
	
	/**
	  Check if this.bitset is a subset of b.bitset
	  @retval true if this.bitset is a subset; false otherwise.
	*/		
	public boolean subset(BitInterface b)
	{
		ByteClass other = (ByteClass) b;
		
		if (positionsSet != null)
			convertRep();
		if (other.positionsSet != null)
			other.convertRep();

		for (int i = 0; i < bsSize; i++) 
		{
			// if bit is set in this, then it better be set in other
			if (get(i) != 0) 
			{
				if (other.get(i) == 0)
				{
					return false;
				}
			}
		}
		return true;
	}

		
	/**
	Clear all the bits in this bitset.	
	*/		
	public void clearAll() 
	{
		lastPosition = 0;
		positionsSet = new int[MAX_POSITIONS_TRACKED];
		// bs = new short[bsSize];
		nBitsSet = 0;
	}
	
	/**
	Clear a bit at a particular position.
	@param num:  The position that should be cleared.	
	*/		
	public void clear(int num) 
	{
		int loc = num >> 3; // divide by 8
		int off = num & 0x7;
		
		// is already clear
		if (get(num) == 0)
			return;
			
		bs[loc] = (short) (bs[loc] & (~(1 << off)));
		nBitsSet--;
	}
	
	/**
	Compute the position that defines the maximal bit set in this Bitset.
	@retval result
	*/		
	public int maxBitSet()
	{
		int max = -1;
		
		if (positionsSet != null) 
		{
			for (int i = 0; i < lastPosition; i++) 
			{
				if (positionsSet[i] > max)
					max = positionsSet[i];
			}
			return max;
		}
		
		for (int i = 0; i < bsSize; i++)
		{
			if (this.get(i) != 0)
				max = i;
		}
		return max;
	}
	

	/**
	Clone this bitset.
	@retval a cloned object.
	*/		
	public Object clone() 
	{
		ByteClass bc = new ByteClass(this.bsSize);
		
		if (this.positionsSet != null) 
		{
			for (int i = 0; i < lastPosition; i++) 
			{
				bc.positionsSet[i] = this.positionsSet[i];
			}
			bc.lastPosition = this.lastPosition;
			bc.nBitsSet = nBitsSet;
			return bc;
		}
		bc.positionsSet = null;
		bc.lastPosition = 0;
		bc.bs = new short[bsSize];
		for (int i = 0; i < bsSize; i++) 
		{
			bc.bs[i] = this.bs[i];
		}
		bc.nBitsSet = nBitsSet;
		return bc;
	}
}

