/**

PROJECT: CS 298

FILE:    DiskReader.java

PURPOSE: Emulates a disk by reading 4K in bulk; the caller consumes one line at a time.

@author 	Aarathi Raghu	

@version 	12/05/2006

@since 		Java 1.4.2

*/	



import java.io.*;
import java.lang.Object.*;
import java.util.*;
import java.lang.*;

/**
 Emulate a disk: Read 4k bytes in bulk; the caller consumes one line at a time.
*/
public class DiskReader 
{
		FileReader fr;
		BufferedReader br;
		int bufPageSize = 4096;
		char bufPage[] = new char[bufPageSize];
		int bufPageStartOff;
		int nread, nbytesInBuf, curPos;
	
		/**
		Constructor
		@param filename:  The file to read data from.
		*/	
		public DiskReader(String filename) throws Exception
		{
			try 
			{
				fr = new FileReader(filename);
				br = new BufferedReader(fr);
			}
			catch (Exception e) 
			{
				e.printStackTrace();
			}
			nread = nbytesInBuf = bufPageStartOff = curPos = 0;
		}
	
		/**
		Read a block of data from input file.
		*/
		
		private void ReadBlock() throws Exception
		{
				// shift things over in case we had previously read part of an integer
				for (bufPageStartOff = 0; curPos < nbytesInBuf; curPos++,bufPageStartOff++) 
				{
					bufPage[bufPageStartOff] = bufPage[curPos];
				}
				// Read a page at a time.
				nread = br.read(bufPage, bufPageStartOff, bufPageSize - bufPageStartOff);
				curPos = 0;
				// # of bytes we have: what we just read + what we had previously
				nbytesInBuf = nread + bufPageStartOff;
		}
	
		/**
		 Satisfy the user-request either from the buffer of data that was read due to
		 a previous call to this function or read a block from disk and satisfy user request.
		 The caller will supply a buffer to hold all the integer values on a given line.
	         @param lineVals: The line of values read from the file.
		*/
		public int ReadLine(int lineVals[]) throws Exception
		{
			int gotLine = -1;
			int curInt = 0;
			
			while (true)
			{
				boolean gotInt = false;
				String intVal = "";
				char ch = 'a'; //something other than ' ' or '\n'.
				
				// Extract an integer based on what we had read in
				while (curPos < nbytesInBuf) 
				{
					ch = bufPage[curPos];
					curPos++;
					if ((ch == ' ') || (ch == '\n'))
					{
						if (intVal != "") 
						{
							gotInt = true;
							break;
						}
						if (ch == '\n')
							break;
						if (ch == ' ') // eat whitespace
							continue;
					}
					intVal = intVal + ch;
				}
				
				if (gotInt)
				{
					// move the pointer foward
					bufPageStartOff = curPos;
					// record the integer
					lineVals[curInt] = Integer.parseInt(intVal, 10);
					// System.out.println("Read: " + lineVals[curInt] + " at pos= " + curInt);
					curInt++;
				}
				// stop if we read a new line
				if (ch == '\n')
				{
					gotLine = 1;
					break;
				}
				// So, we don't have an integer...read more.
				if (!gotInt) 
				{
					// reset curPos to reflect the char's we didn't consume
					curPos = bufPageStartOff;
					ReadBlock();
					if (nread < 0)
						return -1;
					continue;
				}

			}
			if (gotLine > 0)
				return curInt;
			else
				return -1;
		}
}
       
