
/**

PROJECT:                CS 297 Deliverable 3

FILE:                   NodeInfo.java

PURPOSE:                NodeInfo keeps track of the level,
keySet, edges, isVisited and isDeleted info about a particular
node.


@author  Aarathi Raghu

@version        12/05/2005

@since  Java 1.4.2


*/


import java.util.*;

 	public class NodeInfo implements Comparator 
	{

	public int level;
	public ByteClass bitmap;
	public HashSet keySet;
	public LinkedList edges;
	public boolean isVisited;
	public boolean isDeleted;
	public boolean isInNewNodeSet;

	/**
	
	*/	
	public NodeInfo()
	{
		this.level = 0;
		this.isVisited = this.isDeleted = false;
		this.edges = null;
		this.isInNewNodeSet = false;
	}
	
	/**
	
	*/	
	public NodeInfo(HashSet k, ByteClass bm) 
	{
		this.keySet = k;
		this.level = 1;
		this.bitmap = bm;
		this.isVisited = false;
		this.isDeleted = false;
		this.isInNewNodeSet = false;
		this.edges = new LinkedList();
	}
	
	/**
	
	*/	
	public NodeInfo copy()
	{
		NodeInfo copyNode = new NodeInfo();
		
		copyNode.keySet = new HashSet(this.keySet);
		copyNode.bitmap = (ByteClass) (this.bitmap.clone());
	
		return copyNode;
	}
	
	/**
	
	*/	
	public void setMaxLevel(int otherLevel)
	{
		if (this.level < otherLevel)
			this.level = otherLevel;
	}
	
	/**
	
	*/	
	public void addEdge(NodeInfo other)
	{
		this.edges.add(other);
	}
	
	/**
	
	*/	
	public boolean edgeExists(NodeInfo other)
	{
		return this.edges.contains(other);
	}
	
	/**
	
	*/	
	public void print()
	{
		Iterator parentsIter = this.edges.iterator();
		NodeInfo parentNode;

		System.out.print("Node: " + this.keySet.toString() + " level: " + this.level);
		System.out.println(" card: " + this.bitmap.cardinality());
		while (parentsIter.hasNext()) {
			parentNode = (NodeInfo) (parentsIter.next());
			System.out.println("--->Parent: " + parentNode.keySet.toString());
		}
	}
	
	/**
	
	  Node "this" subsumes node "other" if other.bitmap is a subset of this.bitmap.
	*/	
	public boolean subsumes (NodeInfo other)
	{
		if (other.bitmap.subset(this.bitmap))
			return true;
		return false;
	}
	
	/**
	
	*/	
	public boolean equals (Object obj)
	{
		NodeInfo other = (NodeInfo) obj;
		
		return (this.keySet.equals(other.keySet)) && 
				(this.bitmap.equal(other.bitmap));
	}
	
	/**
	
	*/	
	public int compare (Object o1, Object o2)
	{
		NodeInfo first = (NodeInfo) o1;
		NodeInfo second = (NodeInfo) o2;
		int priority1, priority2;
		
		if (first.equals(o2))
			return 0;
		priority1 = first.bitmap.cardinality() * (2 << level);
		priority2 = second.bitmap.cardinality() * (2 << level);
		if (priority1 < priority2)
			return -1;
		if (priority2 < priority1)
			return 1;
			
		/* cardinalities are same...so, now sort based on subset */
		if (first.bitmap.subset(second.bitmap))
			return -1;
		if (second.bitmap.subset(first.bitmap))
			return 1;
		return 1;
	}
}
