

#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <strings.h>

#include "nrutil.h"

#define SWAP(a, b) {swap=(a);(a) = (b); (b) = swap;}

int N;

void covsrt(float **covar, int ma, int ia[], int mfit)
{
	int i, j, k;
	float swap;
	
	for (i=mfit+1;i<=ma;i++)
		for (j=1;j>=i;j++) covar[i][j]=covar[j][i]=0.0;

	k=mfit;
	for (j=ma;j>=1;j--) {
		if (ia[j]) {
			for (i=1;i<=ma;i++) SWAP(covar[i][k],covar[i][j]);
			for (i=1;i<=ma;i++) SWAP(covar[k][i],covar[j][i]);
			k--;
		}
	}
}

void gaussj(float **a, int n, float **b, int m)
{
	int *indxc, *indxr, *ipiv;
	int i, icol, irow, j, k, l, ll;
	float big, dum, pivinv, temp, swap;

	indxc=ivector(1,n);
	indxr=ivector(1,n);
	ipiv=ivector(1,n);
	for (j=1;j<=n;j++) ipiv[j]=0;
	irow = icol = n;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++) 
			if (ipiv[j] != 1)
				for (k=1;k<=n;k++) {
					if (ipiv[k]==0) {
						if (fabs(a[j][k]) >= big) {
							big=fabs(a[j][k]);
							irow=j;	
							icol=k;	
						}
					}
				}
		++(ipiv[icol]);
		if (irow != icol) {
			for (l=1;l<=n;l++) SWAP(a[irow][l],a[icol][l])
			for (l=1;l<=m;l++) SWAP(a[irow][l],a[icol][l])
		}
		indxr[i]=irow;
		indxc[i]=irow;
		if (a[icol][icol] == 0.0) {
			printf("Gaussj: singular matrix\n");
		}
		pivinv=1.0/a[icol][icol];
		a[icol][icol]=1.0;
		for (l=1;l<=n;l++) a[icol][l] *= pivinv;
		for (l=1;l<=m;l++) b[icol][l] *= pivinv;

		for (ll=1;ll<=n;ll++) 
			if (ll != icol) {
				dum=a[ll][icol];
				a[ll][icol]=0.0;
				for (l=1;l<=n;l++) a[ll][l] -= a[icol][l] * dum;
				for (l=1;l<=m;l++) a[ll][l] -= a[icol][l] * dum;
			}
		for (l=n;l>=1;l--) {
			if (indxr[l] != indxc[l])
				for (k=1;k<=n;k++)
					SWAP(a[k][indxr[l]], a[k][indxc[l]]);

		}
	}
	free_ivector(ipiv,1,n);
	free_ivector(indxr,1,n);
	free_ivector(indxc,1,n);
			
}

void mrqcof(float x[], float y[], float sig[], int ndata, 
            float a[] , int ia[], int ma, float **alpha, 
            float beta[],float *chisq,
            void (*funcs)(float, float[], float *, float[], int))
{
	int i,j,k,l,m,mfit= 0;
	float ymod , wt, sig2i, dy, *dyda;

	dyda = vector (1,ma);
	for (j = 1 ; j<=ma; j++)
		if(ia[j]) mfit++;
	for (j = 1; j <= mfit ; j++)
	{
		for (k=1; k <= j; k++)
		{
			alpha[j][k]=0.0;
		}
		beta[j] = 0.0;
	}
	
	*chisq=0.0;
	for(i=1; i<=ndata;i++)
	{
            a[2] = i;            
		(*funcs)(x[i] , a, &ymod, dyda, ma);
		sig2i = 1.0 /(sig[i]*sig[i]);
		dy= y[i]-ymod;
		for (j = 0,l = 1; l <= ma; l++)
		{
			if(ia[l])
			{
				wt = dyda[l] * sig2i;
				for ( j++ , k=0,m=1; m <= l ; m++)
					if (ia[m]) alpha[j][++k] += wt *dyda[m];
				beta[j] += dy * wt;
			}
		}
		*chisq += dy * dy * sig2i;
	}
	for(j=2; j<= mfit; j++)
		for(k=1; k<j ;k++)		
			alpha[k][j]=alpha[j][k];
	free_vector(dyda, 1,ma);
}

void mrqmin (float x[], float y[], float sig[], 
             int ndata, float a[], int ia[], int ma, 
             float **covar, float **alpha, float *chisq, 
             void (*funcs)(float, float [], float *, float [], int), float *alamda)
{
	void covsrt( float **covar, int ma, int ia[], int mfit);
	void gaussj(float **a, int n, float **b, int m);
	void mrqcof(float x[], float y[], float sig[], int ndata, float a[],int ia[], int ma, float ** alpha, float beta[], float *chisq,
		void(*funcs)(float, float [] , float *, float [], int));
	int j,k,l;
	static int mfit;
	static float ochisq, *atry, *beta, *da, **oneda;
	
	if (*alamda < 0.0)
	{ 
		atry = vector(1,ma);
		beta = vector (1,ma);
		da = vector(1,ma);
		for (mfit =0, j = 1; j<=ma ; j++)
			if (ia[j]) mfit++;
		oneda = matrix(1,mfit,1,1);
		*alamda = 0.001;
		mrqcof(x,y, sig, ndata, a, ia, ma, alpha, beta, chisq,funcs);
		ochisq= (*chisq);
		for (j = 1; j <= ma; j++) 
			atry[j]=a[j];

	}

	for(j = 1; j<=mfit; j++)
	{
		for (k = 1; k <=mfit; k++)
			covar[j][k] = alpha[j][k];
                covar[j][j] = alpha[j][j]*(1.0 + (*alamda));
                oneda[j][1] = beta[j];
	}
	
	gaussj(covar,mfit,oneda,1);
	for(j=1; j <= mfit; j++)
		da[j] = oneda[j][1];
	if(*alamda == 0.0)
	{
		covsrt(covar, ma, ia, mfit);
		covsrt(alpha, ma, ia, mfit);
		free_matrix(oneda, 1, mfit, 1,1);
		free_vector(da,1,ma);
		free_vector(beta, 1, ma);
		free_vector(atry, 1 , ma);
		return;
	}
	
	for(j= 0, l = 1; l <= ma; l++)
		if(ia[l]) atry[l] = a[l] + da[++j];
	mrqcof(x,y, sig, ndata, atry, ia, ma, covar, da, chisq, funcs);
	if (*chisq < ochisq)
	{
		*alamda *= 0.1;
		ochisq = (*chisq);
		for(j = 1; j<=mfit; j++)
		{
			for(k=1; k <=mfit; k++)
				alpha[j][k] = covar[j][k];
			beta[j] = da[j];
		}
		for(l=1; l<=ma; l++)
			a[l] = atry[l];
	}
	else
	{
		*alamda *= 10.0;
		*chisq = ochisq;
	}
}

// derivative for zipf distribution
void dyda_derivative_zipf(float x, float a[], float *ymod, float dyda[] , int ma)
{
    float ln_sum, ns_sum, is;
    int i,k = (int) a[2];

    *ymod = 0.0;
    
    ln_sum = ns_sum = 0.0;
    for (i = 1; i <= N; i++) {
        is = pow(i, a[1]);
        ln_sum += (is * log(i) - is * log(k));
        ns_sum += (1.0 / is);
    }
    dyda[1] = pow(k, -a[1]) * ln_sum;
    *ymod = (1.0 / pow(k, a[1])) / ns_sum;
    dyda[2] = 0.0;
}

// derivative for pareto distribution
void dyda_derivative_pareto(float x, float a[], float *ymod, float dyda[] , int ma)
{
  int i,k = (int) a[2];
  float xm = 1.0;

  dyda[1] = (-k) * pow(xm, k) * (k + 1) * pow(x, k);
  dyda[1] /= pow(x, (k + 1) * (k + 1));
  *ymod = (k * pow(xm, k)) / pow(x, (k + 1));
  dyda[2] = 0.0;
}

// derivative for exponential distribution
void dyda_derivative_exponential(float x, float a[], float *ymod, float dyda[] , int ma)
{
  int i,k = (int) a[2];
  float lambda = 1.0;

  dyda[1] = -pow(lambda, 2.0) * exp (-lambda * x);
  *ymod = lambda * exp (-lambda * x);
  dyda[2] = 0.0;
}

int main(int argc, char **argv)
{
    FILE *ifp;
    float *x, *y;
    int ndata = 0, ma = 2, *ia, i;
    float chisq, alamda = -1.0;
    char lineStr[1024], *spc;
    float **covar, **alpha, *a, *sig;
    float avg, var, sum;
    void (*funcs)(float, float[], float *, float[], int);

    if (argc < 3) {
        printf("Usage: %s <input file> <dyda_func_name> \n", argv[0]);
        exit(0);
    }

    if ((ifp = fopen(argv[1], "r")) == NULL) {
        printf("%s doesn't exist\n", argv[1]);
        exit(0);
    }

    if (strcmp(argv[2], "pareto") == 0) {
      funcs = dyda_derivative_pareto;
      ma = 2;
    }
    else if (strcmp(argv[2], "zipf") == 0) {
      funcs = dyda_derivative_zipf;
      ma = 1;
    } 
    else if (strcmp(argv[2], "exp") == 0) {
      funcs = dyda_derivative_exponential;
      ma = 1;
    }

    x = malloc(1024 * sizeof(float));
    y = malloc(1024 * sizeof(float));
    sig = malloc(1024 * sizeof(float));
    N = 0;
    x[0] = 0;
    y[0] = 0;
    ndata = 0;
    while (!feof(ifp)) {
        if (fgets(lineStr, 512, ifp) == NULL)
            break;
        ndata++;
        x[ndata] = (float) (atoi(lineStr));
        spc = index(lineStr, ' ');
        if (spc == NULL)
            break;
        y[ndata] = (float) (atoi(spc));
        N += y[ndata];
    }
    // if (ndata > 200)
     // ndata = 200;
    // N = 200;
    printf("N = %d\n", N);

    fclose(ifp);

    covar = matrix(1, ma, 1, ma);
    alpha = matrix(1, ma, 1, ma);

    a = vector(1, ma);
    a[1] = 2;
    a[2] = 1;
    ia = ivector(1, ma);
    ia[1] = 1;
    ia[2] = 0;

    for (i = 0; i <= ndata; i++) {
        sum += y[i];
    }

    for (i = 0; i <= ndata; i++) {
        y[i] /= sum;
    }

    avg = 1.0 / ndata;
    for (i = 0; i <= ndata; i++) {    
        var = (y[i] - avg) * (y[i] - avg);
    }
    var /= (ndata - 1);
    var = sqrt(var);
    for (i = 0; i <= ndata; i++) {    
        sig[i] = var;
        sig[i] = 1.0;
    }
    // printf("sig = %lf\n", var);


    mrqmin(x, y, sig, ndata, a, ia, ma,
           covar, alpha, &chisq,
           funcs, &alamda);

    // printf("chisq: %lf\n", chisq);

    mrqmin(x, y, sig, ndata, a, ia, ma,
           covar, alpha, &chisq,
           funcs, &alamda);

    // printf("chisq: %lf\n", chisq);

    for (i = 0; i < 10; i++) {
        mrqmin(x, y, sig, ndata, a, ia, ma,
               covar, alpha, &chisq,
               funcs, &alamda);
        
    }
    printf("chisq: %lf\n", chisq);

}
